import logging

from peek_abstract_chunked_index.private.server.controller.ACIProcessorQueueControllerABC import (
    ACIProcessorQueueControllerABC,
    ACIProcessorQueueBlockItem,
)
from peek_abstract_chunked_index.private.server.controller.ACIProcessorStatusNotifierABC import (
    ACIProcessorStatusNotifierABC,
)
from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import (
    ACIProcessorQueueTupleABC,
)
from peek_core_search._private.server.client_handlers.ClientSearchObjectChunkUpdateHandler import (
    ClientSearchObjectChunkUpdateHandler,
)
from peek_core_search._private.server.controller.StatusController import (
    StatusController,
)
from peek_core_search._private.storage.EncodedSearchObjectChunk import (
    EncodedSearchObjectChunk,
)
from peek_core_search._private.storage.SearchObject import SearchObject
from peek_core_search._private.storage.SearchObjectCompilerQueue import (
    SearchObjectCompilerQueue,
)

logger = logging.getLogger(__name__)


class _Notifier(ACIProcessorStatusNotifierABC):
    def __init__(self, adminStatusController: StatusController):
        self._adminStatusController = adminStatusController

    def setProcessorStatus(self, state: bool, queueSize: int):
        self._adminStatusController.status.searchObjectCompilerQueueStatus = (
            state
        )
        self._adminStatusController.status.searchObjectCompilerQueueSize = (
            queueSize
        )
        self._adminStatusController.notify()

    def addToProcessorTotal(self, delta: int):
        self._adminStatusController.status.searchObjectCompilerQueueProcessedTotal += (
            delta
        )
        self._adminStatusController.notify()

    def setProcessorError(self, error: str):
        self._adminStatusController.status.searchObjectCompilerQueueLastError = (
            error
        )
        self._adminStatusController.notify()


class SearchObjectChunkCompilerQueueController(ACIProcessorQueueControllerABC):
    QUEUE_ITEMS_PER_TASK = 10
    POLL_PERIOD_SECONDS = 1.000

    QUEUE_BLOCKS_MAX = 20
    QUEUE_BLOCKS_MIN = 4

    WORKER_TASK_TIMEOUT = 60.0

    _logger = logger
    _QueueDeclarative: ACIProcessorQueueTupleABC = SearchObjectCompilerQueue
    _VacuumDeclaratives = (
        SearchObjectCompilerQueue,
        SearchObject,
        EncodedSearchObjectChunk,
    )

    def __init__(
        self,
        dbSessionCreator,
        statusController: StatusController,
        clientSearchObjectUpdateHandler: ClientSearchObjectChunkUpdateHandler,
    ):
        ACIProcessorQueueControllerABC.__init__(
            self, dbSessionCreator, _Notifier(statusController)
        )
        self._clientSearchObjectUpdateHandler: ClientSearchObjectChunkUpdateHandler = (
            clientSearchObjectUpdateHandler
        )

    def _sendToWorker(self, block: ACIProcessorQueueBlockItem):
        from peek_core_search._private.worker.tasks.SearchObjectChunkCompilerTask import (
            compileSearchObjectChunk,
        )

        return compileSearchObjectChunk.delay(block.itemsEncodedPayload)

    def _processWorkerResults(self, results):
        self._clientSearchObjectUpdateHandler.sendChunks(results)

    def _dedupeQueueSql(self, lastFetchedId: int, dedupeLimit: int):
        return """
                 with sq_raw as (
                    SELECT "id", "chunkKey"
                    FROM core_search."SearchObjectCompilerQueue"
                    WHERE id > %(id)s
                    LIMIT %(limit)s
                ), sq as (
                    SELECT min(id) as "minId", "chunkKey"
                    FROM sq_raw
                    GROUP BY  "chunkKey"
                    HAVING count("chunkKey") > 1
                )
                DELETE
                FROM core_search."SearchObjectCompilerQueue"
                     USING sq sq1
                WHERE core_search."SearchObjectCompilerQueue"."id" != sq1."minId"
                    AND core_search."SearchObjectCompilerQueue"."id" > %(id)s
                    AND core_search."SearchObjectCompilerQueue"."chunkKey" = sq1."chunkKey"

            """ % {
            "id": lastFetchedId,
            "limit": dedupeLimit,
        }



"""add full exclude terms

Peek Plugin Database Migration Script

Revision ID: 79609b0e25a5
Revises: 3dba609610b6
Create Date: 2023-03-01 23:21:19.427518

"""

# revision identifiers, used by Alembic.
revision = "79609b0e25a5"
down_revision = "3dba609610b6"
branch_labels = None
depends_on = None

import sqlalchemy as sa
from alembic import op


def upgrade():
    op.add_column(
        "ExcludeSearchString",
        sa.Column("partial", sa.Boolean(), nullable=True),
        schema="core_search",
    )

    op.add_column(
        "ExcludeSearchString",
        sa.Column("full", sa.Boolean(), nullable=True),
        schema="core_search",
    )


def downgrade():
    op.drop_column("ExcludeSearchString", "partial", schema="core_search")
    op.drop_column("ExcludeSearchString", "full", schema="core_search")



from typing import List, Dict

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search._private.storage.SearchObjectTypeTuple import (
    SearchObjectTypeTuple,
)
from peek_core_search.tuples.ImportSearchObjectRouteTuple import (
    ImportSearchObjectRouteTuple,
)
from vortex.Tuple import Tuple, addTupleType, TupleField


@addTupleType
class SearchResultObjectTuple(Tuple):
    """Import Search Object

    This tuple is used by other plugins to load objects into the search index.

    """

    __tupleType__ = searchTuplePrefix + "SearchResultObjectTuple"

    # The id of the object this search result is for
    id: int = TupleField()

    #:  The unique key for this object
    key: str = TupleField()

    #:  The type of this object
    objectType: SearchObjectTypeTuple = TupleField()

    #:  The details to index.
    # Do not include "key", it will be indexed anyway.
    # The key of the property will match of create a new "SearchProperty"
    properties: Dict[str, str] = TupleField({})

    #:  The color
    routes: List[ImportSearchObjectRouteTuple] = TupleField([])

    #:  The rank of the search result
    rank: int = TupleField()



import logging
from typing import Union

from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

from peek_core_search._private.storage.ExcludeSearchStringTable import (
    ExcludeSearchStringTable,
)
from peek_core_search._private.tuples.ExcludeSearchStringsTuple import (
    ExcludeSearchStringsTuple,
)

logger = logging.getLogger(__name__)


class ExcludeSearchStringsTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        session = self._ormSessionCreator()
        try:
            partials = []
            fulls = []
            for exclude in session.query(ExcludeSearchStringTable):
                if exclude.partial:
                    partials.append(exclude.term)

                if exclude.full:
                    fulls.append(exclude.term)

            tuple_ = ExcludeSearchStringsTuple()
            tuple_.excludedPartialSearchTerms = partials
            tuple_.excludedFullSearchTerms = fulls

            # Create the vortex message
            return (
                Payload(filt, tuples=[tuple_])
                .makePayloadEnvelope()
                .toVortexMsg()
            )

        finally:
            session.close()



import logging

from sqlalchemy import String

from peek_abstract_chunked_index.private.tuples.ACIProcessorQueueTupleABC import (
    ACIProcessorQueueTupleABC,
)
from peek_core_search._private.PluginNames import searchTuplePrefix
from sqlalchemy import Column, BigInteger
from sqlalchemy import Integer
from vortex.Tuple import Tuple, addTupleType

from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class SearchObjectCompilerQueue(
    Tuple, DeclarativeBase, ACIProcessorQueueTupleABC
):
    __tablename__ = "SearchObjectCompilerQueue"
    __tupleType__ = searchTuplePrefix + "SearchObjectCompilerQueueTable"

    id = Column(BigInteger, primary_key=True, autoincrement=True)
    chunkKey = Column(String, primary_key=True)

    @classmethod
    def sqlCoreLoad(cls, row):
        return SearchObjectCompilerQueue(id=row.id, chunkKey=row.chunkKey)

    @property
    def ckiUniqueKey(self):
        return self.chunkKey



import logging

from peek_core_search._private.storage.DeclarativeBase import loadStorageTuples
from peek_core_search._private.tuples import loadPrivateTuples
from peek_core_search.tuples import loadPublicTuples
from peek_plugin_base.worker.plugin_worker_entry_hook_abc import (
    PluginWorkerEntryHookABC,
)

logger = logging.getLogger(__name__)


class WorkerEntryHook(PluginWorkerEntryHookABC):
    def load(self):
        loadPrivateTuples()
        loadStorageTuples()
        loadPublicTuples()
        logger.debug("loaded")

    def start(self):
        logger.debug("started")

    def stop(self):
        logger.debug("stopped")

    def unload(self):
        logger.debug("unloaded")



from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy import orm
from vortex.Tuple import Tuple
from vortex.Tuple import addTupleType

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search._private.storage.DeclarativeBase import DeclarativeBase


@addTupleType
class SearchPropertyTuple(DeclarativeBase, Tuple):
    __tupleType__ = searchTuplePrefix + "SearchPropertyTuple"
    __tablename__ = "SearchProperty"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    title = Column(String, nullable=False)
    order = Column(Integer, nullable=False, server_default="0")

    showOnResult = Column(Boolean, nullable=False, server_default="1")
    showInHeader = Column(Boolean, nullable=False, server_default="0")

    __table_args__ = (
        Index("idx_SearchProp_name", name, unique=True),
        Index("idx_SearchProp_title", title, unique=True),
    )

    @orm.reconstructor
    def __init__(self, **kwargs):
        Tuple.__init__(self, **kwargs)



from twisted.trial import unittest

from peek_core_search._private.worker.tasks.KeywordSplitter import (
    prepareExcludedPartialTermsForFind,
)
from peek_core_search._private.worker.tasks.KeywordSplitter import (
    splitPartialKeywords,
    splitFullKeywords,
    _splitFullTokens,
    twoCharTokens,
)


class KeywordSplitterTest(unittest.TestCase):
    def testDuplicates(self):
        self.assertEqual({"smith"}, _splitFullTokens("smith smith"))

    def test_twoCharTokens(self):
        self.assertEqual(set(), twoCharTokens("smith smith"))

        self.assertEqual({"^to$"}, twoCharTokens(splitFullKeywords("two to")))

        self.assertEqual(
            {"^to"}, twoCharTokens(splitPartialKeywords([], "two to"))
        )

    def testFullKeywordSplit(self):
        self.assertEqual({"^smith$"}, splitFullKeywords(set(), "smith"))
        self.assertEqual(
            {"^zorro-reyner$"}, splitFullKeywords(set(), "ZORRO-REYNER")
        )
        self.assertEqual({"^34534535$"}, splitFullKeywords(set(), "34534535"))

        self.assertEqual({"^and$"}, splitFullKeywords(set(), "and"))
        self.assertEqual({"^to$"}, splitFullKeywords(set(), "to"))
        self.assertEqual({"^to$"}, splitFullKeywords(set(), "to"))

        self.assertEqual(
            {"^milton$", "^unit$", "^22$"},
            splitFullKeywords(set(), "Milton Unit 22"),
        )

        self.assertEqual({"^unit$"}, splitFullKeywords(set(), "Unit A"))

        self.assertEqual({"^unit$"}, splitFullKeywords(set(), "Unit 1"))

        self.assertEqual(
            {"^trans$", "^66kv$", "^b3$", "^cb$", "^ats$"},
            splitFullKeywords(set(), "ATS B3 TRANS 66KV CB"),
        )

    def testFullKeywordSplitWithExcludes(self):
        self.assertEqual(
            {"^66kv$", "^b3$", "^cb$"},
            splitFullKeywords({"trans", "ats"}, "ATS B3 TRANS 66KV CB"),
        )

        self.assertEqual(
            {"^66kv$", "^b3$", "^cb$", "^trans$"},
            splitFullKeywords({"tran", "ats"}, "ATS B3 TRANS 66KV CB"),
        )

        self.assertEqual(
            {"^milton$", "^22$"}, splitFullKeywords({"unit"}, "Milton Unit 22")
        )

    def testPartialKeywordSplit(self):
        self.assertEqual(
            {"^smi", "mit", "ith"}, splitPartialKeywords([], "smith")
        )

        self.assertEqual(
            {
                "^zor",
                "orr",
                "rro",
                "ro-",
                "o-r",
                "-re",
                "rey",
                "eyn",
                "yne",
                "ner",
            },
            splitPartialKeywords([], "ZORRO-REYNER"),
        )
        self.assertEqual(
            {"^345", "535", "453", "534", "345"},
            splitPartialKeywords([], "34534535"),
        )

        self.assertEqual({"^and"}, splitPartialKeywords([], "and"))

        self.assertEqual({"^to"}, splitPartialKeywords([], "to"))

        self.assertEqual({"a55", "^ha5"}, splitPartialKeywords([], "ha55"))

        self.assertEqual(
            {"^mil", "ilt", "lto", "ton", "^uni", "nit", "^22"},
            splitPartialKeywords([], "Milton Unit 22"),
        )

        self.assertEqual(
            {"^mil", "ill", "lls", "^un", "^no"},
            splitPartialKeywords([], "mills un no"),
        )

        self.assertEqual(
            {"^uni", "nit", "^22"}, splitPartialKeywords([], "Unit 22")
        )

        self.assertEqual({"^uni", "nit"}, splitPartialKeywords([], "Unit 1"))

        self.assertEqual({"^uni", "nit"}, splitPartialKeywords([], "A Unit"))

        self.assertEqual({"^uni", "nit"}, splitPartialKeywords([], "2 Unit"))

        self.assertEqual(
            {"^ats", "^cb", "^b3", "^66k", "6kv", "^tra", "ran", "ans"},
            splitPartialKeywords([], "ATS B3 TRANS 66KV CB"),
        )

        self.assertEqual(
            {"^col", "^lin", "ins"}, splitPartialKeywords([], "COL LINS")
        )

        self.assertNotEqual(
            splitPartialKeywords([], "COLLINS"),
            splitPartialKeywords([], "COL LINS"),
        )

    def test_splitPartialWithExcludes(self):
        # Test to ensure we can remove useless tokens.
        self.assertEqual(
            {"^123", "234", "345", "456"},
            splitPartialKeywords(["AliAS-".lower()], "ALIAS-123456"),
        )

        self.assertEqual(
            {"^123", "234", "345", "456", "567"},
            splitPartialKeywords(
                ["AliAS-".lower(), "wEc".lower()], "WEC1234567"
            ),
        )

    def test_preparePartialExcludedTermsForFind(self):
        self.assertEqual(
            ["123456", "12345", "1234", "123", "1"],
            prepareExcludedPartialTermsForFind(["1234", "123", "123456", "1"]),
        )



import hashlib
import json
import logging
from base64 import b64encode
from collections import defaultdict
from datetime import datetime
from typing import List, Dict

import pytz
from sqlalchemy import select

from vortex.Payload import Payload

from peek_core_search._private.storage.EncodedSearchObjectChunk import (
    EncodedSearchObjectChunk,
)
from peek_core_search._private.storage.SearchObject import SearchObject
from peek_core_search._private.storage.SearchObjectCompilerQueue import (
    SearchObjectCompilerQueue,
)
from peek_plugin_base.worker.task_db_conn import TaskDbConn
from peek_plugin_base.worker.task import addPeekWorkerTask

logger = logging.getLogger(__name__)

""" Search Index Compiler

Compile the search indexes

1) Query for queue
2) Process queue
3) Delete from queue
"""


@addPeekWorkerTask()
def compileSearchObjectChunk(payloadEncodedArgs: bytes) -> dict[str, str]:
    """Compile Search Index Task


    :param payloadEncodedArgs: An encoded payload containing the queue tuples.
    :returns: A list of grid keys that have been updated.
    """
    argData = Payload().fromEncodedPayload(payloadEncodedArgs).tuples
    queueItems = argData[0]
    queueItemIds: List[int] = argData[1]

    chunkKeys = list(set([i.chunkKey for i in queueItems]))

    queueTable = SearchObjectCompilerQueue.__table__
    compiledTable = EncodedSearchObjectChunk.__table__
    lastUpdate = datetime.now(pytz.utc).isoformat()

    startTime = datetime.now(pytz.utc)

    engine = TaskDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()
    try:
        logger.debug(
            "Staring compile of %s queueItems in %s",
            len(queueItems),
            (datetime.now(pytz.utc) - startTime),
        )

        # Get Model Sets

        total = 0
        existingHashes = _loadExistingHashes(conn, chunkKeys)
        encKwPayloadByChunkKey = _buildIndex(chunkKeys)
        chunksToDelete = []

        lastUpdateByChunkKey = {}

        inserts = []
        for (
            chunkKey,
            searchIndexChunkEncodedPayload,
        ) in encKwPayloadByChunkKey.items():
            m = hashlib.sha256()
            m.update(searchIndexChunkEncodedPayload)
            encodedHash = b64encode(m.digest()).decode()

            # Compare the hash, AND delete the chunk key
            if chunkKey in existingHashes:
                # At this point we could decide to do an update instead,
                # but inserts are quicker
                if encodedHash == existingHashes.pop(chunkKey):
                    continue

            lastUpdateByChunkKey[str(chunkKey)] = lastUpdate

            chunksToDelete.append(chunkKey)
            inserts.append(
                dict(
                    chunkKey=chunkKey,
                    encodedData=searchIndexChunkEncodedPayload,
                    encodedHash=encodedHash,
                    lastUpdate=lastUpdate,
                )
            )

        # Add any chnuks that we need to delete that we don't have new data for, here
        chunksToDelete.extend(list(existingHashes))

        if chunksToDelete:
            # Delete the old chunks
            conn.execute(
                compiledTable.delete().where(
                    compiledTable.c.chunkKey.in_(chunksToDelete)
                )
            )

        if inserts:
            newIdGen = TaskDbConn.prefetchDeclarativeIds(
                SearchObject, len(inserts)
            )
            for insert in inserts:
                insert["id"] = next(newIdGen)

        if inserts:
            conn.execute(compiledTable.insert(), inserts)

        conn.execute(
            queueTable.delete().where(queueTable.c.id.in_(queueItemIds))
        )

        transaction.commit()

        logger.debug(
            "Compiled %s SearchObjects, %s missing, in %s",
            len(inserts),
            len(chunkKeys) - len(inserts),
            (datetime.now(pytz.utc) - startTime),
        )

        total += len(inserts)

        logger.info(
            "Compiled and Committed %s EncodedSearchObjectChunks in %s",
            total,
            (datetime.now(pytz.utc) - startTime),
        )

        return lastUpdateByChunkKey

    except Exception as e:
        transaction.rollback()
        logger.warning(e)  # Just a warning, it will retry
        # logger.exception(e)
        raise

    finally:
        conn.close()


def _loadExistingHashes(conn, chunkKeys: List[str]) -> Dict[str, str]:
    compiledTable = EncodedSearchObjectChunk.__table__

    results = conn.execute(
        select(compiledTable.c.chunkKey, compiledTable.c.encodedHash).where(
            compiledTable.c.chunkKey.in_(chunkKeys)
        )
    ).fetchall()

    return {result[0]: result[1] for result in results}


def _buildIndex(chunkKeys) -> Dict[str, bytes]:
    session = TaskDbConn.getDbSession()

    try:
        indexQry = (
            session.query(
                SearchObject.chunkKey, SearchObject.id, SearchObject.packedJson
            )
            .filter(SearchObject.chunkKey.in_(chunkKeys))
            .filter(SearchObject.packedJson != None)
            .order_by(SearchObject.id)
            .yield_per(1000)
            .all()
        )

        # Create the ChunkKey -> {id -> packedJson, id -> packedJson, ....]
        packagedJsonByObjIdByChunkKey = defaultdict(dict)

        for item in indexQry:
            packagedJsonByObjIdByChunkKey[item.chunkKey][
                item.id
            ] = item.packedJson

        encPayloadByChunkKey = {}

        # Sort each bucket by the key
        for chunkKey, packedJsonById in packagedJsonByObjIdByChunkKey.items():
            tuples = json.dumps(packedJsonById, sort_keys=True)

            # Create the blob data for this index.
            # It will be searched by a binary sort
            encPayloadByChunkKey[chunkKey] = (
                Payload(tuples=tuples).toEncodedPayload().encode()
            )

        return encPayloadByChunkKey

    finally:
        session.close()



import hashlib
import json
import logging
from _collections import defaultdict
from base64 import b64encode
from datetime import datetime
from functools import cmp_to_key
from typing import List, Dict

import pytz
from sqlalchemy import select

from vortex.Payload import Payload

from peek_core_search._private.storage.EncodedSearchIndexChunk import (
    EncodedSearchIndexChunk,
)
from peek_core_search._private.storage.SearchIndex import SearchIndex
from peek_core_search._private.storage.SearchIndexCompilerQueue import (
    SearchIndexCompilerQueue,
)
from peek_plugin_base.worker.task_db_conn import TaskDbConn
from peek_plugin_base.worker.task import addPeekWorkerTask

logger = logging.getLogger(__name__)

""" Search Index Compiler

Compile the search indexes

1) Query for queue
2) Process queue
3) Delete from queue
"""


@addPeekWorkerTask()
def compileSearchIndexChunk(payloadEncodedArgs: bytes) -> dict[str, str]:
    """Compile Search Index Task


    :param payloadEncodedArgs: An encoded payload containing the queue tuples.
    :returns: A list of grid keys that have been updated.
    """
    argData = Payload().fromEncodedPayload(payloadEncodedArgs).tuples
    queueItems = argData[0]
    queueItemIds: List[int] = argData[1]

    chunkKeys = list(set([i.chunkKey for i in queueItems]))

    queueTable = SearchIndexCompilerQueue.__table__
    compiledTable = EncodedSearchIndexChunk.__table__
    lastUpdate = datetime.now(pytz.utc).isoformat()

    startTime = datetime.now(pytz.utc)

    engine = TaskDbConn.getDbEngine()
    conn = engine.connect()
    transaction = conn.begin()
    try:
        logger.debug(
            "Staring compile of %s queueItems in %s",
            len(queueItems),
            (datetime.now(pytz.utc) - startTime),
        )

        # Get Model Sets

        total = 0
        existingHashes = _loadExistingHashes(conn, chunkKeys)
        encKwPayloadByChunkKey = _buildIndex(conn, chunkKeys)
        chunksToDelete = []

        lastUpdateByChunkKey = {}

        inserts = []
        for (
            chunkKey,
            searchIndexChunkEncodedPayload,
        ) in encKwPayloadByChunkKey.items():
            m = hashlib.sha256()
            m.update(searchIndexChunkEncodedPayload)
            encodedHash = b64encode(m.digest()).decode()

            # Compare the hash, AND delete the chunk key
            if chunkKey in existingHashes:
                # At this point we could decide to do an update instead,
                # but inserts are quicker
                if encodedHash == existingHashes.pop(chunkKey):
                    continue

            lastUpdateByChunkKey[str(chunkKey)] = lastUpdate

            chunksToDelete.append(chunkKey)
            inserts.append(
                dict(
                    chunkKey=chunkKey,
                    encodedData=searchIndexChunkEncodedPayload,
                    encodedHash=encodedHash,
                    lastUpdate=lastUpdate,
                )
            )

        # Add any chnuks that we need to delete that we don't have new data for, here
        chunksToDelete.extend(list(existingHashes))

        if chunksToDelete:
            # Delete the old chunks
            conn.execute(
                compiledTable.delete().where(
                    compiledTable.c.chunkKey.in_(chunksToDelete)
                )
            )

        if inserts:
            newIdGen = TaskDbConn.prefetchDeclarativeIds(
                SearchIndex, len(inserts)
            )
            for insert in inserts:
                insert["id"] = next(newIdGen)

        transaction.commit()
        transaction = conn.begin()

        if inserts:
            conn.execute(compiledTable.insert(), inserts)

        logger.debug(
            "Compiled %s SearchIndexes, %s missing, in %s",
            len(inserts),
            len(chunkKeys) - len(inserts),
            (datetime.now(pytz.utc) - startTime),
        )

        total += len(inserts)

        conn.execute(
            queueTable.delete().where(queueTable.c.id.in_(queueItemIds))
        )

        transaction.commit()
        logger.info(
            "Compiled and Committed %s EncodedSearchIndexChunks in %s",
            total,
            (datetime.now(pytz.utc) - startTime),
        )

        return lastUpdateByChunkKey

    except Exception as e:
        transaction.rollback()
        # logger.warning(e)  # Just a warning, it will retry
        logger.exception(e)
        raise

    finally:
        conn.close()


def _loadExistingHashes(conn, chunkKeys: List[str]) -> Dict[str, str]:
    compiledTable = EncodedSearchIndexChunk.__table__

    results = conn.execute(
        select(compiledTable.c.chunkKey, compiledTable.c.encodedHash).where(
            compiledTable.c.chunkKey.in_(chunkKeys)
        )
    ).fetchall()

    return {result[0]: result[1] for result in results}


def _buildIndex(conn, chunkKeys) -> Dict[str, bytes]:
    indexTable = SearchIndex.__table__

    results = conn.execute(
        select(
            indexTable.c.chunkKey,
            indexTable.c.keyword,
            indexTable.c.propertyName,
            indexTable.c.objectId,
        ).where(indexTable.c.chunkKey.in_(chunkKeys))
    )

    encKwPayloadByChunkKey = {}

    # Create the SearchTerm -> SearchProperty -> [objectIds, objectIds, ....] structure
    objIdsByPropByKwByChunkKey = defaultdict(
        lambda: defaultdict(lambda: defaultdict(list))
    )

    for item in results:
        (
            objIdsByPropByKwByChunkKey[item.chunkKey][item.keyword][
                item.propertyName
            ].append(item.objectId)
        )

    def _sortSearchIndex(o1, o2):
        if o1[0] < o2[0]:
            return -1
        if o1[0] > o2[0]:
            return 1
        if o1[1] < o2[1]:
            return -1
        if o1[1] > o2[1]:
            return 1
        return 0

    # Sort each bucket by the key
    for chunkKey, objIdsByPropByKw in objIdsByPropByKwByChunkKey.items():
        compileSearchIndexChunks = []

        for keyword, objIdsByProp in objIdsByPropByKw.items():
            if len(objIdsByProp) > 1000:
                logger.error(
                    "Too many items in bucket for keyword %s",
                    keyword,
                    len(objIdsByProp),
                )

            for propertyName, objectIds in objIdsByProp.items():
                compileSearchIndexChunks.append(
                    [keyword, propertyName, json.dumps(list(sorted(objectIds)))]
                )

        compileSearchIndexChunks.sort(key=cmp_to_key(_sortSearchIndex))

        # Create the blob data for this index.
        # It will be searched by a binary sort
        encKwPayloadByChunkKey[chunkKey] = (
            Payload(tuples=compileSearchIndexChunks).toEncodedPayload().encode()
        )

    return encKwPayloadByChunkKey
